<?php

namespace Drupal\webt\Controller;

use Drupal\Component\Utility\Html;
use Drupal\Core\Controller\ControllerBase;
use Drupal\webt\translation_engines\etranslation\EtranslationService;
use Drupal\webt\translation_engines\etranslation\EtranslationUtils;
use Drupal\webt\TranslationHelper;
use Drupal\webt\WebtConstInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

/**
 * Controller that handles eTranslation async responses.
 */
class EtranslationCallbackController extends ControllerBase {
  /**
   * The translation manager service.
   *
   * @var \Drupal\webt\TranslationManager
   */
  protected $translationManager;

  /**
   * The logger service.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * Entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Shared temp storage.
   *
   * @var \Drupal\Core\TempStore\SharedTempStore
   */
  protected $store;

  /**
   * Constructs a new class instance.
   *
   * @param \Drupal\webt\TranslationManager $translationManager
   *   The language manager service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   */
  public function __construct($translationManager, $entityTypeManager, $shared_temp_store_factory) {
    $this->translationManager = $translationManager;
    $this->logger             = $this->getLogger(WebtConstInterface::WEBT_LOGGER);
    $this->entityTypeManager  = $entityTypeManager;
    $this->store              = $shared_temp_store_factory->get('webt');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('webt.translation_manager'),
      $container->get('entity_type.manager'),
      $container->get('tempstore.shared'),
    );
  }

  /**
   * Receive & process async translation response from eTranslation.
   *
   * If this is a translation for an entity and timeout has been reached before,
   * also parse and save translation.
   *
   * @param string $id
   *   The translation request ID.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The HTTP request object.
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   The HTTP response.
   *
   * @throws \Symfony\Component\HttpKernel\Exception\BadRequestHttpException
   *   If the request ID is invalid.
   */
  public function onResponse($id, Request $request) {
    $key  = EtranslationUtils::getStoreKey($id);
    $data = $this->store->get($key);

    if (!$data) {
      throw new BadRequestHttpException('Invalid request ID.');
    }

    $request_id         = Html::escape($request->get('request-id'));
    $translated_content = base64_decode(file_get_contents('php://input'));

    $this->logger->info("Received response from eTranslation (@request_id)", [
      '@request_id' => $request_id,
    ]);

    if ($data['translation'] === EtranslationUtils::$timeoutValue) {
      $this->logger->info("Received late translation (@request_id)", [
        '@request_id' => $request_id,
      ]);

      $this->store->delete($key);

      if (!empty($data['entity_id']) && !empty($data['entity_type'])) {
        $entity_id = $data['entity_id'];
        $entity_type = $data['entity_type'];
        $entity = $this->translationManager->loadEntityByIdAndType($entity_id, $entity_type);

        if ($entity) {
          $lang = Html::escape($request->get('target-language'));

          $this->logger->info(
            "Adding translation to @lang for entity @entity_type @entity_id (@request_id)", [
              '@lang' => $lang,
              '@entity_id' => $entity_id,
              '@entity_type' => $entity_type,
              '@request_id' => $request_id,
            ]
          );

          $this->processEntityTranslation($entity, $translated_content, $lang);
        }
      }
    }
    else {
      $data['translation'] = $translated_content;
      $this->store->set($key, $data);
    }

    return new Response();
  }


  /**
   * Process async error response from eTranslation.
   *
   * Logs error and marks translation
   * as completed with error.
   *
   * @param string $id
   *   WEB-T request ID.
   * @param Symfony\Component\HttpFoundation\Request $request
   *   HTTP request object.
   *
   * @return Symfony\Component\HttpFoundation\Response
   *   Response
   */
  public function onError($id, Request $request){
    $key  = EtranslationUtils::getStoreKey($id);
    $data = $this->store->get($key);
    if(!$data){
      throw new BadRequestHttpException('Invalid request ID.');
    }
    $this->logger->error("eTranslation response error (@request_id). Please make sure eTranslation API is working as expected.", ['@request_id' => $request_id]);
    $data['translation'] = EtranslationUtils::$errorValue;
    $this->store->set($key, $data);
    return new Response();
  }

  /**
   * Parse translation response and insert entity translation.
   *
   * @param [type] $entity
   *   Original entity to translate.
   * @param [type] $translated_content
   *   Translation received from eTranslation service.
   * @param [type] $target_language
   *   Language of translation received.
   */
  private function processEntityTranslation($entity, $translated_content, $target_language) {
    $target_langcode     = $this->getLangcodeByLanguage($target_language);
    $translation_service = $this->translationManager->translationService;
    if ($translation_service instanceof EtranslationService) {
      $raw_translations = $translation_service->etranslationResponseToTranslationArray($translated_content);

      $translatable_fields  = $this->translationManager->getTranslatableFields($entity)['translatable'];
      $original_values      = $this->translationManager->extractTranslatableValues($entity, $translatable_fields);
      
      $translatable_strings = TranslationHelper::extractTranslatableHtmlValues($original_values);
      $translations         = TranslationHelper::mergeTranslatedHtmlValues($original_values, $translatable_strings, $raw_translations);
      
      if (count($original_values) === count($translations)) {
        $this->translationManager->translateEntity($entity->id(), $entity->getEntityTypeId(), $target_langcode, FALSE, $translations);
      }
    }
  }

  /**
   * In languages added to the website find langcode for given language.
   *
   * @param string $language
   *   Requested language e.g. 'en'.
   *
   * @return string
   *   Langcode or null if langcode not found.
   */
  private function getLangcodeByLanguage($language) {
    $languages = $this->languageManager()->getLanguages();
    $language_lowercase = strtolower($language);

    foreach ($languages as $langcode => $value) {
      if (str_starts_with($langcode, $language_lowercase)) {
        return $langcode;
      }
    }
    $this->logger->error("Could not find langcode for language '@language'. Please check if the language code of this language is correct!", ['@language' => $language_lowercase]);
    return NULL;
  }

}
