<?php

namespace Drupal\webt;

use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\webt\Model\ProcessingResponse;
use Drupal\webt\Model\StringType;
use Drupal\webt\Model\TranslationStatus;

/**
 * Service that creates/updates/deletes translated versions of source items.
 */
class TranslationManager {

  use LoggerChannelTrait;

  /**
   * Machine translation service that translates strings.
   *
   * @var Drupal\webt\translation_engines\AbstractTranslationService
   */
  public $translationService;

  /**
   * Drupal config factory.
   *
   * @var Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * Language manager service.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;

  /**
   * Locale storage.
   *
   * @var \Drupal\locale\StringDatabaseStorage
   */
  protected $localeStorage;
  /**
   * Config storage.
   *
   * @var \Drupal\Core\Config\CachedStorage
   */
  protected $configStorage;
  /**
   * Entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;
  /**
   * Default language.
   *
   * @var \Drupal\Core\Language\LanguageDefault
   */
  protected $languageDefault;

  /**
   * Logger service.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;
  
  /**
   * Bundle info service.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfo;
   */
  protected $entityTypeBundleInfo;

  /**
   * Content translation manager service.
   *
   * @var \Drupal\content_translation\ContentTranslationManager
   */
  protected $contentTranlationManager;

  /**
   * Field types that contain text.
   *
   * @var array
   */
  private $field_text_types = ['string', 'string_long', 'text', 'text_long', 'text_with_summary'];

  /**
   * Entity type IDs of entities that must not be translated
   *
   * @var array
   */
  private $non_translatable_entity_types = ['user'];

  /**
   * Field names of fields that must not be translated
   *
   * @var array
   */
  private $non_translatable_fields = [
    'path' => ['langcode'],
    'moderation_state' => ['value'],
  ];

  /**
   * Initializes TranslationManager.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param \Drupal\webt\translation_engines\generic\GenericMTService $genericMtService
   *   The generic machine translation service.
   * @param \Drupal\webt\translation_engines\etranslation\EtranslationService $etranslationService
   *   The eTranslation service.
   * @param \Drupal\Core\Language\LanguageManagerInterface $languageManager
   *   The language manager service.
   * @param \Drupal\locale\StringDatabaseStorage $localeStorage
   *   The locale storage service.
   * @param \Drupal\Core\Config\CachedStorage $configStorage
   *   The config storage service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   * @param \Drupal\Core\Language\LanguageDefault $languageDefault
   *   The default language service.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfo $entityTypeBundleInfo
   *   Bundle info service.
   * @param \Drupal\Core\Entity\ContentTranslationManagerInterface $content_tranlation_manager
   *   The content translation manager service.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    $genericMtService,
    $etranslationService,
    $languageManager,
    $localeStorage,
    $configStorage,
    $entityTypeManager,
    $languageDefault,
    $entityTypeBundleInfo,
    $content_tranlation_manager
  ) {
    $this->configFactory = $config_factory;
    $engine = $config_factory->get('webt.settings')->get('mt_engine');
    $this->translationService = $engine !== 'etranslation' ? $genericMtService : $etranslationService;
    $this->languageManager = $languageManager;
    $this->localeStorage = $localeStorage;
    $this->configStorage = $configStorage;
    $this->entityTypeManager = $entityTypeManager;
    $this->languageDefault = $languageDefault;
    $this->entityTypeBundleInfo = $entityTypeBundleInfo;
    $this->contentTranlationManager = $content_tranlation_manager;
    $this->logger = $this->getLogger(WebtConstInterface::WEBT_LOGGER);
  }

  /**
   * Create translations for given entities in given target languages.
   *
   * @param array $entities
   *   Entities to be translated.
   * @param string $type
   *   Entity type needed for progress bar message updates.
   * @param string[] $target_langcodes
   *   Target language codes.
   * @param callable $progress_callback
   *   Progress callback function.
   * @param bool $skip_if_translated
   *   By default function skips already translated entities.
   *   Retranslates them if $skip_if_translated is false.
   *
   * @return int
   *   Value of ProcessingResponse
   */
  public function translateEntities(array $entities, array $target_langcodes, callable $progress_callback, bool $skip_if_translated = TRUE): int {
    $field_data            = [];
    $translations          = [];
    $entities_to_translate = [];

    $to_create = 0;
    $created   = 0;

    // Collect translatable entities, their fields and field values
    foreach ($entities as $entity) {
      $entity_id = spl_object_id($entity);
      $src_lang  = $entity->language()->getId();
  
      if (!isset($field_data[$entity_id])) {
        $fields                   = $this->getTranslatableFields($entity);
        $values                   = $this->extractTranslatableValues($entity, $fields['translatable']);
        $field_data[$entity_id]   = ['fields' => $fields, 'translatable_values' => $values];
      }

      foreach ($target_langcodes as $trg_lang) {
        if ($trg_lang === $src_lang || ($skip_if_translated && $entity->hasTranslation($trg_lang))) {
          continue;
        }

        $entities_to_translate[$src_lang][$trg_lang][] = $entity;
        $translations[$src_lang][$trg_lang][]          = $field_data[$entity_id]['translatable_values'];
        $to_create++;
      }
    }

    // Translate and apply translations.
    foreach ($translations as $src_lang => $targets) {
      foreach ($targets as $trg_lang => $values) {
        $flat_values = array_merge(...$values);

        // Machine translate translatable field values
        $translations = [];
        if (!empty($flat_values)) {
          $translatable_strings = TranslationHelper::extractTranslatableHtmlValues($flat_values);
          $response = $this->translationService->translate($src_lang, $trg_lang, $translatable_strings, $progress_callback);
          if (empty($response->translations) || TranslationStatus::FULL_RESULT !== $response->status) {
            continue;
          }
          $translations = TranslationHelper::mergeTranslatedHtmlValues($flat_values, $translatable_strings, $response->translations);
        }

        // Insert translated entities.
        $offset = 0;
        foreach ($entities_to_translate[$src_lang][$trg_lang] as $entity) {
          $entity_id           = spl_object_id($entity);
          $entity_fields       = $field_data[$entity_id];
          $value_count         = count($entity_fields['translatable_values']);
          $entity_translations = array_slice($translations, $offset, $value_count);
          $offset             += $value_count;

          $this->entityCreateTranslation($entity, $trg_lang, $entity_fields['fields'], $entity_translations);
          $created++;
        }
      }
    }

    return $this->getProcessingResponse($created, $to_create);
  }

  /**
   * Translate entity from specified source language to target language.
   *
   * @param int $entityId
   *   Entity ID
   * @param string $entityTypeId
   *   Entity type ID
   * @param string[] $trg_lang
   *   Target language code
   * @param boolean $skip_if_translation_exists
   *   Whether to skip translation if translation already exists
   * @param string[] $translations
   *   If provided, uses given string translations, skips machine translation.
   */
  public function translateEntity($entityId, $entityTypeId, $trg_lang, $skip_if_translation_exists = FALSE, $translations = NULL) {
    $entity = $this->loadEntityByIdAndType($entityId, $entityTypeId);
    if (!$entity) {
      return;
    }    
    $src_lang = $entity->language()->getId();

    if (!$skip_if_translation_exists || !$entity->hasTranslation($trg_lang)) {
      $entity_fields       = $this->getTranslatableFields($entity);
      $translatable_fields = $entity_fields['translatable'];
      if (count($translatable_fields) > 0) {
        $this->logger->info("Translating entity (@src_lang-@trg_lang)", ['@src_lang' => $src_lang, '@trg_lang' => $trg_lang]);
  
        if (!$translations) {
          $translatable_values  = $this->extractTranslatableValues($entity, $translatable_fields);
          $strings_to_translate = TranslationHelper::extractTranslatableHtmlValues($translatable_values);
          $translation_response = $this->translationService->translate($src_lang, $trg_lang, $strings_to_translate, NULL, $entity);
          if (TranslationStatus::FULL_RESULT !== $translation_response->status) {
            return;
          }
          $translations = TranslationHelper::mergeTranslatedHtmlValues($translatable_values, $strings_to_translate, $translation_response->translations);
        }
        $this->entityCreateTranslation($entity, $trg_lang, $entity_fields, $translations);
      }
    }
  }

  /**
   * Delete translations of entities in specified languages.
   *
   * @param \Drupal\Core\Entity\Entity[] $entities
   *   Entities to delete translations of.
   * @param string[] $langcodes
   *   Language codes of translations which need to be deleted.
   *
   * @return int
   *   Value of ProcessingResponse
   */
  public function deleteTranslationsOfEntities($entities, $langcodes) {
    $response = ProcessingResponse::$fullyProcessed;
    foreach ($entities as $entity) {
      $source_langcode = $entity->language()->getId();
      foreach ($langcodes as $langcode) {
        if ($langcode === $source_langcode) {
          continue;
        }
        if ($entity->hasTranslation($langcode)) {          
          try {
            $entity->removeTranslation($langcode);
            $entity->save();
          } catch(\Exception $e) { 
            $this->logger->error('Could not delete entity [ID=@id, message: @message]', ['@id' => $entity->id(), '@message' => $e->getMessage(),]);
            $response = ProcessingResponse::$partiallyProcessed;
          }
        }
      }
    }
    return $response;
  }

  /**
   * Retrieves translation languages for entity
   *
   * @param \Drupal\Core\Entity\Entity $entity
   *   Entity to translate.
   * @return string[]
   *   List of languages
   */
  public function getTranslationLanguagesForEntity($entity) {
    $source_langcode       = $entity->language()->getId();
    $langcodes             = $this->languageManager->getLanguages();
    $langcodes_list        = array_keys($langcodes);
    $translation_languages = [$source_langcode => []];

    foreach ($langcodes_list as $langcode) {
      if ($langcode === $source_langcode || !$this->translationService->isLanguageSupported($langcode)) {
        continue;
      }
      $translation_languages[$source_langcode][] = $langcode;
    }
    return $translation_languages;
  }

  /**
   * Translate given UI strings and save translations.
   *
   * @param \Drupal\locale\TranslationString[] $translatable_strings
   *   Translatable UI strings.
   * @param string $src_lang
   *   Source language of UI strings.
   * @param string $trg_lang
   *   Target language.
   * @param callable $progress_callback
   *   Progress callback function.
   *
   * @return int
   *   Value of ProcessingResponse
   */
  public function translateUiStrings($translatable_strings, $src_lang, $trg_lang, $progress_callback) {
    $values               = array_map(
    function ($s) {
      return $s->source;
    },
            $translatable_strings
    );
    $strings_to_translate = TranslationHelper::extractTranslatableHtmlValues($values);

    $translation_response = $this->translationService->translate($src_lang, $trg_lang, $strings_to_translate, $progress_callback);
    $translations = TranslationHelper::mergeTranslatedHtmlValues($values, $strings_to_translate, $translation_response->translations, TRUE);
    $count        = count($translations);

    for ($i = 0; $i < $count; $i++) {
      $translated_string = $translations[$i];
      $translation       = $this->localeStorage->createTranslation(
      [
        'lid'         => $translatable_strings[$i]->lid,
        'language'    => $trg_lang,
        'translation' => $translated_string,
      ]
      );
      $translation->save();
    }
    return $this->getProcessingResponse($count, count($values));
  }

  /**
   * Get list of all untranslated UI strings for a language.
   *
   * @param string $langcode
   *   Translation language code.
   *
   * @return \Drupal\locale\TranslationString[]
   *   Untranslated UI strings
   */
  public function getUntranslatedUiStrings($langcode) {
    $existing_translations = $this->localeStorage->getTranslations(['language' => $langcode]);

    $untranslated_strings = array_filter(
    $existing_translations,
    function ($s) {
      $has_source      = isset($s->source) && !empty($s->source);
      $has_translation = isset($s->translation);
      return $has_source && !$has_translation;
    }
    );
    return array_values($untranslated_strings);
  }

  /**
   * Delete all UI translations for given languages.
   *
   * @param string[] $langcodes
   *   Language codes to delete translations of.
   *
   * @return int
   *   Value of ProcessingResponse
   */
  public function deleteUiTranslations($langcodes) {
    $source_langcode = $this->getDefaultLanguageCode();

    foreach ($langcodes as $langcode) {
      if ($langcode === $source_langcode) {
        continue;
      }
      $this->localeStorage->deleteTranslations(['language' => $langcode]);
    }
    return ProcessingResponse::$fullyProcessed;
  }

  /**
   * Translate given configuration strings to a language.
   *
   * @param array $config_strings
   *   Translatable configuration string items.
   * @param string $langcode
   *   Language code to translate to.
   * @param callable $progress_callback
   *   Progress callback function.
   * @return int
   *   Value of ProcessingResponse
   */
  public function translateConfigStrings($config_strings, $langcode, $progress_callback) {
    // Prepare source value list for translation.
    $source_strings = [];
    foreach ($config_strings as $item) {
      array_walk_recursive(
      $item,
      function ($p) use (&$source_strings) {
        if (is_string($p)) {
          $source_strings[] = $p;
        }
      }
      );
    }
    $offset = 0;
    if (!empty($source_strings)) {
      // translate.
      $source_langcode      = $this->getDefaultLanguageCode();      
      $strings_to_translate = TranslationHelper::extractTranslatableHtmlValues($source_strings);
      $translation_response = $this->translationService->translate($source_langcode, $langcode, $strings_to_translate, $progress_callback);
      $translations = TranslationHelper::mergeTranslatedHtmlValues($source_strings, $strings_to_translate, $translation_response->translations);

      // Replace source values with translations & save.
      if (!empty($translations) && TranslationStatus::FULL_RESULT === $translation_response->status) {
        $collection = $this->configStorage->createCollection("language.$langcode");
        foreach ($config_strings as $key => $value) {
          $translated_data = $value;
          array_walk_recursive(
          $translated_data,
          function (&$v) use ($translations, &$offset) {
            if (is_string($v)) {
              $v = $translations[$offset];
              $offset++;
            }
          }
          );
          $collection->write($key, $translated_data);
        }
      }
    }

    return $this->getProcessingResponse($offset, count($source_strings));
  }

  /**
   * Gets all untranslated cofiguration strings for a language.
   *
   * @param string $langcode
   *   Language code.
   *
   * @return array
   *   Untranslated configuration strings
   */
  public function getUntranslatedConfigStrings($langcode) {
    $translatable_items    = $this->getTranslatableConfigStrings();
    $translated_item_names = $this->configStorage->createCollection("language.$langcode")->listAll();

    // Get untranslated config item names.
    $missing_items = array_filter(
    $translatable_items,
    function ($name) use ($translated_item_names) {
      return !in_array($name, $translated_item_names, TRUE);
    },
            ARRAY_FILTER_USE_KEY
    );

    return $missing_items;
  }

  /**
   * Get all translatable configuration strings (in source language).
   *
   * @return array
   *   Translatable config strings
   */
  private function getTranslatableConfigStrings() {
    $source_item_names  = $this->configFactory->listAll();
    $config_items       = $this->configFactory->loadMultiple($source_item_names);
    $translatable_items = [];

    // Identify translatable configuration items and their fields.
    foreach ($config_items as $item) {
      $original = $item->getOriginal();
      if (isset($original['langcode'])) {
        $translatable_fields = $this->getTranslatableConfigFields($original);
        if (!empty($translatable_fields)) {
          $translatable_items[$item->getName()] = $translatable_fields;
        }
      }
    }

    return $translatable_items;
  }

  /**
   * Get translatable properties & their values for a configuration string item.
   *
   * @param mixed $config_item
   *   Configuration string item.
   *
   * @return array
   *   Translatable config fields
   */
  private function getTranslatableConfigFields($config_item) {
    $translatable_fields = [];
    // Keys that usually hold translatable values.
    $search_keys = ['label', 'name', 'description', 'message'];
    // Arrays that can contain $search_keys.
    $container_keys = ['settings'];

    foreach ($search_keys as $key) {
      // Check 1st level values.
      if (isset($config_item[$key]) && !is_array($config_item[$key]) && !empty($config_item[$key])) {
        if ('name' === $key && in_array('label', array_keys($translatable_fields), TRUE)) {
          // Do not translate 'name' if 'label' already present.
          // Workaround for drupal/admin/config/media/image-styles translation.
          continue;
        }
        $translatable_fields[$key] = $config_item[$key];
      }
      // Check 2nd level values.
      foreach ($container_keys as $container_key) {
        if (isset($config_item[$container_key]) && isset($config_item[$container_key][$key]) &&
        strlen($config_item[$container_key][$key]) > 0) {
          if (!isset($translatable_fields[$container_key])) {
            $translatable_fields[$container_key] = [];
          }
          $translatable_fields[$container_key][$key] = $config_item[$container_key][$key];
        }
      }
    }

    return $translatable_fields;
  }

  /**
   * Delete all configuration string translations for given languages.
   *
   * @param string[] $langcodes
   *   Target language codes.
   *
   * @return int
   *   Value of ProcessingResponse
   */
  public function deleteConfigTranslations($langcodes) {
    $source_langcode = $this->getDefaultLanguageCode();

    foreach ($langcodes as $langcode) {
      if ($langcode === $source_langcode) {
        continue;
      }
      $this->configStorage->createCollection("language.$langcode")->deleteAll();
    }
    return ProcessingResponse::$fullyProcessed;
  }

  /**
   * Finds and groups translatable and referenced fields.
   *
   * @param EntityInterface $entity
   *   The entity to check.
   * @return array
   *   A multidimensional list of fields.
   */
  function getTranslatableFields($entity) {
    $fields = [
      'translatable' => array(),
      'referenced'   => array(),
    ];
  
    if (!$entity instanceof FieldableEntityInterface) {
      return $fields;
    }

    $non_translatable_field_names = array_keys($this->non_translatable_fields);  
    foreach ($entity->getFieldDefinitions() as $field_name => $field_definition) {  
      if ($field_definition->isTranslatable()) {
        switch ($field_definition->getType()) {
          case 'entity_reference':
            $fields['referenced'][$field_name] = [];
            break;
          case 'language':
            // Skip language types as they will be different for translated entity
            break;
          default:
            $item_definition = $field_definition->getItemDefinition();
            foreach ($item_definition->getPropertyDefinitions() as $property_name => $property_definition) {
              if (!$property_definition->isComputed() && in_array($property_definition->getDataType(), $this->field_text_types)) {

                if (in_array($field_name, $non_translatable_field_names) && in_array($property_name, $this->non_translatable_fields[$field_name])) {
                  if (!isset($fields['referenced'][$field_name])) {
                    $fields['referenced'][$field_name] = [];
                  }
                  $fields['referenced'][$field_name][] = $property_name;
                  continue;
                }

                if (!isset($fields['translatable'][$field_name])) {
                  $fields['translatable'][$field_name] = [];
                }
                $fields['translatable'][$field_name][] = $property_name;

              }
            }
            break;
        }
      }
    }
    return $fields;
  }

  /**
   * Checks if entity needs retranslation.
   *
   * @param \Drupal\Core\Entity\Entity $entity
   *   Entity.
   *
   * @return bool
   *   True if translatable content of entity has changed. False otherwise
   */
  public function entityTranslatableContentChanged($entity) {
    $translatable_fields = $this->getTranslatableFields($entity);

    foreach ($translatable_fields['translatable'] as $parent => $fields) {
      foreach ($fields as $field) {
        if (!$entity->original || ($entity->$parent && $entity->original->$parent && $entity->$parent->$field != $entity->original->$parent->$field)) {
          return TRUE;
        }
      }
    }

    foreach($translatable_fields['referenced'] as $parent => $fields ) {
      if ($entity->hasField($parent) && is_callable([$entity->$parent, 'referencedEntities']) && $entity->$parent->referencedEntities() !== $entity->original->$parent->referencedEntities()) {
        return TRUE;
      }
    }
    return FALSE;
  }

  /**
   * Get translation progress percentage for given content type and language.
   *
   * @param string $type
   *   Content type.
   * @param string $langcode
   *   Target language code.
   *
   * @return float
   *   Translation progress in percents
   */
  public function getTranslationProgressPercents($type, $langcode) {
    $source_language  = $this->getDefaultLanguageCode();
    $source_count     = 0;
    $translated_count = 0;
    switch ($type) {
      case StringType::UI:
        if ($source_language === $langcode) {
          return 100;
        }
        $source_count     = count($this->localeStorage->getStrings());
        $translated_count = $source_count - count($this->getUntranslatedUiStrings($langcode));
        break;

      case StringType::CONFIG:
        if ($source_language === $langcode) {
          return 100;
        }
        else {
          $source_strings   = $this->getTranslatableConfigStrings();
          $source_count     = count($source_strings);
          $translated_count = $source_count - count($this->getUntranslatedConfigStrings($langcode));
          break;
        }
      default:
        $entity_types = $this->getTranslatableEntityTypes();
        foreach($entity_types as $entity_type_id => $bundles) {
          $source_entities = $this->getEntitiesByTypeAndBundles($entity_type_id, $bundles);
          $translated_entities = array_filter(
            $source_entities,
            function ($e) use ($langcode) {
              return $e->language()->getId() === $langcode || $e->hasTranslation($langcode);
            }
          );
          $source_count     += count($source_entities);
          $translated_count += count($translated_entities);
        }
        break;
    }
    return $source_count > 0 ? (float) 100 * $translated_count / $source_count : 0;
  }

  /**
   * Get default language code of the website.
   *
   * @return string
   *   Default langcode
   */
  public function getDefaultLanguageCode() {
    return $this->languageDefault->get()->getId();
  }

  /**
   * Create translated version of an entity.
   *
   * @param \Drupal\Core\Entity\Entity $entity
   *   Entity to translate.
   * @param string $trg_lang
   *   Target language code.
   * @param array $translation_fields
   *   Translatable fields for this entity.
   * @param array $translation_values
   *   Translated values of translatable fields.
   */
  private function entityCreateTranslation($entity, $trg_lang, $entity_fields, $translation_values) {
    $translation_fields = $entity_fields['translatable'];
    $translation_count  = count($translation_values);
    $source_value_count = count($this->extractTranslatableValues($entity, $translation_fields));
    if ($source_value_count === $translation_count) {
      if ($entity->hasTranslation($trg_lang)) {
        $entity->removeTranslation($trg_lang);
      }

      $field_translations = [];      
      // Copy reference fields from original entity
      foreach( array_keys($entity_fields['referenced']) as $field ) {
        $field_translations[$field] = $entity->get($field)->getValue();
      }

      // Copy translation fields and update values with translations
      $offset = 0;
      foreach ($translation_fields as $parent => $fields) {
        $field_translations[$parent] = $entity->get($parent)->getValue();
        if (empty(array_keys($field_translations[$parent]))) {
          $offset += count($fields);
          continue;
        }
        foreach(array_keys($field_translations[$parent]) as $key) {
          foreach ($fields as $field) {
            $translated_string = $translation_values[$offset++];
            if ($parent === 'path' && $field === 'alias' && strlen($translated_string) > 0) {
              $field_translations[$parent][$key][$field] = '/' . urlencode(str_replace('/', '', $translated_string));
            } else {
              $field_translations[$parent][$key][$field] = $translated_string;
            }
          }
        }
      }

      $translation = $entity->addTranslation($trg_lang, $field_translations);

      if ($entity->hasField('status')) {
        $translation->status->value = $entity->status->value;
      }

      if ($entity->hasField('content_translation_source')) {
        $translation->content_translation_source->value = $entity->langcode->value;
      }

      $translation->save();
    }
  }

  /**
   * Get processing response status based on translation and reference count.
   *
   * @param int $translation_count
   *   Translated string value count.
   * @param int $reference_value_count
   *   Original (reference) string value count.
   *
   * @return int
   *   Value of ProcessingResponse
   */
  private function getProcessingResponse($translation_count, $reference_value_count) {
    if (0 === $reference_value_count) {
      return ProcessingResponse::$alreadyProcessed;
    }
    elseif ($translation_count === $reference_value_count) {
      return ProcessingResponse::$fullyProcessed;
    }
    elseif (0 === $translation_count && 0 < $reference_value_count) {
      return ProcessingResponse::$processingError;
    }
    else {
      return ProcessingResponse::$partiallyProcessed;
    }
  }

  /**
   * Get translatable values from an entity using list of translatable fields.
   *
   * @param \Drupal\Core\Entity\Entity $entity
   *   Entity.
   * @param array $translatable_fields
   *   Translatable fields of given entity.
   *
   * @return string[]
   *   Translatable values
   */
  public function extractTranslatableValues($entity, $translatable_fields) {
    $translatable_values = [];
    foreach ($translatable_fields as $parent => $fields) {
      foreach ($fields as $field) {
        $translatable_values[] = $entity->$parent->$field ?? '';
      }
    }
    return $translatable_values;
  }

  /**
   * Retrieve entity by ID.
   *
   * @param [type] $id
   *   Identifier of entity.
   * @param [type] $type
   *   Type of entity.
   *
   * @return \Drupal\Core\Entity\Entity
   *   Entity or null if enitity was not found
   */
  public function loadEntityByIdAndType($id, $type) {
    $storage = $this->entityTypeManager->getStorage($type);
    $entity  = $storage->load($id);
    if ($entity) {
      return $entity;
    }
    else {
      $this->logger->error("Could not load entity by ID (@entity_id). Please make sure such entity exists!", ['@entity_id' => $id]);
      return NULL;
    }
  }

  /**
   * Retrieves entity types that are translatable and have translation enabled.
   *
   * @return array
   */
  public function getTranslatableEntityTypes() {
    $entityTypeManager             = $this->entityTypeManager;
    $entity_types                  = $entityTypeManager->getDefinitions();       
    $non_translatable_entity_types = $this->non_translatable_entity_types; 

    // Filter translatable entities.
    $translatable_entity_types = array_filter(
      $entity_types, 
      function($entity_type, $entity_type_id) use ($entityTypeManager, $non_translatable_entity_types) {
        $translationPossible = $entity_type->isTranslatable();
        $translationAllowed  = !in_array($entity_type_id, $non_translatable_entity_types);
        return $translationPossible && $translationAllowed;
      },
      ARRAY_FILTER_USE_BOTH
    );

    // Filter translatable bundles of translatable entites.
    $translation_enabled_types = [];
    foreach ($translatable_entity_types as $entity_type_id => $entity_type) {
      $bundles = $this->entityTypeBundleInfo->getBundleInfo($entity_type_id);
      if (empty($bundles)) {
        $translation_enabled_types[$entity_type_id] = [];
        continue;
      }
      foreach (array_keys($bundles) as $bundle_name) {
        if ($this->contentTranlationManager->isEnabled($entity_type_id, $bundle_name)) {
          if (!isset($translation_enabled_types[$entity_type_id])) {
            $translation_enabled_types[$entity_type_id] = [];
          }
          $translation_enabled_types[$entity_type_id][] = $bundle_name;
        }
      }
    }
    return $translation_enabled_types;
  }

  /**
   * Retrieves entities by provided type ID and bundle names
   *
   * @param string   $entity_type_id Entity type ID
   * @param string[] $bundles Bundle names
   * @return object[]
   */
  public function getEntitiesByTypeAndBundles($entity_type_id, $bundles) {
    $storage = $this->entityTypeManager->getStorage($entity_type_id);
    if(empty($bundles)) {
      return $storage->loadMultiple();
    }

    $entity_type = $this->entityTypeManager->getDefinition($entity_type_id);
    $bundle_key  = $entity_type->getKey('bundle');    
    if ($bundle_key) {
      return $storage->loadByProperties([$bundle_key => $bundles]);
    } else {
      return $storage->loadMultiple();
    }
  }

  /**
   * Checks if WEB-T should translate given entity
   *
   * @param object $entity Entity.
   * @return boolean
   */
  public function isEntityTranslatable($entity) {
    $translatable_types = $this->getTranslatableEntityTypes();
    $entity_type_id     = $entity->getEntityTypeId();
    $type_translatable  = array_key_exists($entity_type_id, $translatable_types);
    if (!$type_translatable) {
      return FALSE;
    }
    $bundle_translatable = empty($translatable_types[$entity_type_id]) || in_array($entity->bundle(), $translatable_types[$entity_type_id]);
    $is_original         = array_key_exists('default_langcode', $entity->toArray()) && $entity->default_langcode && $entity->default_langcode[0]->value;
    return $bundle_translatable && $is_original;
  }
}
