<?php

namespace Drupal\webt\translation_engines;

use Drupal\Core\Http\ClientFactory;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\webt\Model\TranslationResponse;
use Drupal\webt\Model\TranslationStatus;
use Drupal\webt\WebtConstInterface;

/**
 * Translation service base class.
 */
abstract class AbstractTranslationService {

  use LoggerChannelTrait;
  use StringTranslationTrait;

  /**
   * HTTP client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;
  /**
   * Whether to retry entity translation on error.
   *
   * @var bool
   */
  protected $retryOnEntityTranslationError = TRUE;
  /**
   * Char limit when to split translatable strings into separate request.
   *
   * @var int
   */
  protected $maxCharsPerRequest = 60000;
  /**
   * Drupal configuration.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * Logger service.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * Initializes service.
   *
   * @param \Drupal\Core\Http\ClientFactory $http_client_factory
   *   HTTP client factory.
   */
  public function __construct(ClientFactory $http_client_factory) {
    $this->httpClient = $http_client_factory->fromOptions(['timeout' => 120]);
    $this->config     = \Drupal::config('webt.settings');
    $this->logger     = $this->getLogger(WebtConstInterface::WEBT_LOGGER);
  }

  /**
   * Translation request function implemented by each MT provider.
   *
   * @param string $from
   *   Source langcode.
   * @param string $to
   *   Target langcode.
   * @param string[] $values
   *   Translatable XML strings.
   * @param Drupal\Core\Entity\EntityInterface $entity
   *   Related entity.
   *
   * @return string[]
   *   Translations
   */
  abstract protected function sendTranslationRequest($from, $to, $values, $entity);

  /**
   * Language direction request function implemented by each MT provider.
   *
   * @return Psr\Http\Message\ResponseInterface
   *   Response
   */
  abstract protected function sendLanguageDirectionRequest();

  /**
   * Checks if MT authorization data is provided.
   *
   * @return bool
   *   True if MT provider data is configured, false otherwise
   */
  abstract public function isAuthorized();

  /**
   * Gets languages supported by MT API from cache or by making an HTTP request.
   *
   * @return array
   *   MT provider supported languages
   */
  public function getSupportedLanguages() {
    $key         = 'language_directions';
    $config      = \Drupal::getContainer()->get('config.factory')->getEditable('webt.settings');
    $json_string = $config->get($key);
    if (!$json_string) {
      $response = $this->sendLanguageDirectionRequest();
      if ($response && 200 === $response->getStatusCode()) {
        $json_string = $this->mapLanguageDirectionResponseBody($response->getBody());
        $config->set($key, $json_string);
        $config->save();
      }
      else {
        $result = [];
        $config->set($key, $result);
        $config->save();
        return $result;
      }
    }
    $obj = json_decode($json_string);
    return $obj->languageDirections ?? [];
  }

  /**
   * Checks if given langcode is supported by MT provider.
   *
   * @param string $langcode
   *   Language code.
   *
   * @return bool
   *   True, if language is supported, false otherwise
   */
  public function isLanguageSupported($langcode) {
    $lang                = substr($langcode, 0, 2);
    $supported_languages = $this->getSupportedLanguages();
    foreach ($supported_languages as $engine) {
      if ($engine->trgLang === $lang) {
        return TRUE;
      }
    }
    return FALSE;
  }

  /**
   * Translates given string values.
   *
   * @param string $from
   *   Source langcode.
   * @param string $to
   *   Target langcode.
   * @param string[] $values
   *   String array of translatable values.
   * @param callable $progress_callback
   *   Progress callback function for pre-translation.
   * @param Drupal\Core\Entity\EntityInterface $entity
   *   Related entity.
   *
   * @return string[]
   *   Translations
   */
  public function translate($from, $to, $values, $progress_callback = NULL, $entity = NULL) {
    $current_char_len    = 0;
    $request_value_lists = [];
    $current_list        = [];
    $lang_from           = substr($from, 0, 2);
    $lang_to             = substr($to, 0, 2);

    // Split into multiple arrays not exceeding max char limit.
    foreach ($values as $value) {
      $chars = strlen($value);
      if ($current_char_len + $chars > $this->maxCharsPerRequest) {
        $request_value_lists[] = $current_list;
        $current_list          = [$value];
        $current_char_len      = $chars;
      }
      else {
        $current_list[]    = $value;
        $current_char_len += $chars;
      }
    }
    if (!empty($current_list)) {
      $request_value_lists[] = $current_list;
    }

    // Send requests and merge results together.`
    $full_result   = [];
    $request_count = count($request_value_lists);

    $percentage_completed = 0;
    for ($i = 0; $i < $request_count; $i++) {
      $this->progressCallback($progress_callback, $percentage_completed, $lang_from, $lang_to);

      $translatable_values = $request_value_lists[$i];
      $this->logger->debug("Sending @lang_from-@lang_to translation request @iteration/@total_iterations (@total_strings strings)",
        [
          '@lang_from' => $lang_from,
          '@lang_to' => $lang_to,
          '@iteration' => $i + 1,
          '@total_iterations' => $request_count,
          '@total_strings' => count($translatable_values),
        ]
      );
      $translations = $this->sendTranslationRequest($lang_from, $lang_to, $translatable_values, $entity);

      if (!$translations || empty($translations)) {
        $retries_left = 2;
        while ($retries_left > 0 && (!$entity || $this->retryOnEntityTranslationError)) {
          $this->progressCallback($progress_callback, $percentage_completed, $lang_from, $lang_to);
          $this->logger->debug("Translation request failed, retrying... (retries left: @retries_left)", ['@retries_left' => $retries_left]);
          $translations = $this->sendTranslationRequest($lang_from, $lang_to, $translatable_values, $entity);
          $retries_left--;
        }

        // Do not continue if one of requests fail.
        if (!$translations || empty($translations)) {
          if ($request_count > 1) {
            $this->logger->error('One of translation requests failed. Please check if machine translation provider API is working or reduce the "Max characters per single request" value (in Translation provider configuration) and try again!');
          }
          return new TranslationResponse($full_result, empty($full_result) ? TranslationStatus::NO_RESULT : TranslationStatus::PARTIAL_RESULT);
        }
      }
      // Merge with previous translations.
      $full_result = array_merge($full_result, $translations);
      $percentage_completed += 100 / $request_count;
    }
    $this->progressCallback($progress_callback, 100, $lang_from, $lang_to);
    return new TranslationResponse($full_result, TranslationStatus::FULL_RESULT);
  }

  /**
   * Makes a language direction request to check if connection is successful.
   *
   * @return bool
   *   True, if MT API request is successful, false otherwise
   */
  public function testConnection() {
    $response = $this->sendLanguageDirectionRequest();
    if (!$response) {
      $this->logger->error('Response object is null. Please check if MT service can be reached.');
      return FALSE;
    }
    $content_type = isset($response->getHeaders()['Content-Type']) ? $response->getHeaders()['Content-Type'][0] : NULL;
    if ($content_type && !str_contains($content_type, 'application/json')) {
      $this->logger->error("Invalid content type '@content_type' in language direction response!", ['@content_type' => $content_type]);
      return FALSE;
    }
    return 200 === $response->getStatusCode();
  }

  /**
   * Maps language direction response to Generic API JSON response.
   *
   * @param string $body
   *   Response body.
   *
   * @return string
   *   JSON body
   */
  protected function mapLanguageDirectionResponseBody($body) {
    return (string) $body;
  }

  /**
   * Call progress callback function with progress info
   *
   * @param callable $callback
   *   Callback function
   * @param float $percentage
   *   Percentage completed.
   * @param string $from
   *   Source language.
   * @param string $to
   *   Target language.
   *
   * @return string|null
   *   Pre-translation request ID or null
   */
  private function progressCallback($callback, $percentage, $from, $to) {
    if ($callback) {
      $callback($percentage, $from, $to);
    }
  }

}
